#!/usr/bin/env python
# encoding: utf-8
"""
manufacturerspecific.py

Created by Graeme Sutherland on 2008-06-23.
Copyright (c) 2008 TrickleStar Limited. All rights reserved.
"""

import sys
import os
import unittest
from zwconstants import *


known_product_types = {
40: ["USB Stick", "400ZW-AM", "Z-Wave USB Stick Mac OSX Widget, English, 8 Control Points"],
}

known_product_ids = {
}

class ManufacturerSpecific(object):
    """
    Hold, convert, decode and display Z-Wave device versions
    """
    def __init__(self):
        self.set_initial_ms()
    
    def set_ms(self, manuf_id, prod_type_id, prod_id):
        self.manuf_id = manuf_id
        self.prod_type_id = prod_type_id
        self.prod_id = prod_id
    
    def set_ms_from_report(self, rc):
        if len(rc)>=6:
            self.set_ms(rc[0:2], rc[2:4], rc[4:6])
        else:
            self.set_initial_ms()
    
    def set_initial_ms(self):
        self.set_ms([0x0, 0x0], [0x0,0x0], [0x0,0x0])
    
    def get_ms(self):
        """ returns array of bytes containing manufacture specific info """
        return self.manuf_id + self.prod_type_id + self.prod_id
    
    def get_manufacturer_for_display(self):
        if self.is_tricklestar():
            return 'TrickleStar'
        elif self.is_null:
            return 'Unknown'
        else:
            return hex((self.manuf_id[0]*0x100) + self.manuf_id[1])
    
    def get_product_type_for_display(self):
        if self.is_null():
            return 'Unknown'
        ptid = (self.prod_type_id[0] * 0x100) + self.prod_type_id[1]
        if known_product_types.has_key(ptid):
            return known_product_types[ptid]
        else:
            return hex(ptid)
    
    def get_product_id_for_display(self):
        if self.is_null():
            return 'Unknown'
        pid = (self.prod_id[0] * 0x100) + self.prod_id[1]
        if known_product_ids.has_key(pid):
            return known_product_ids[pid]
        else:
            return hex(pid)
        
    
    def is_tricklestar(self):
        return self.manuf_id == MANUF_TRICKLESTAR
    
    def is_null(self):
        return self.manuf_id == [0x0,0x0]
