#!/usr/bin/env python
# encoding: utf-8
"""
zwclasses.py

Created by Graeme Sutherland on 2008-04-25.
Copyright (c) 2008 TrickleStar Limited. All rights reserved.
"""

import sys
import os
import unittest
from zwclassresolver import *
from zwtransport import *
from manufacturerspecific import ManufacturerSpecific


# Device classes

class BASIC_TYPE_SLAVE(ZWClass):
    zid = 0x03
    name = 'BASIC_TYPE_SLAVE'
    short_name = 'slave'
    man_classes = []

ZWBasicDeviceClasses.register(BASIC_TYPE_SLAVE)

class BASIC_TYPE_ROUTING_SLAVE(ZWClass):
    zid = 0x04
    name = 'BASIC_TYPE_ROUTING_SLAVE'
    short_name = 'routing_slave'
    man_classes = []

ZWBasicDeviceClasses.register(BASIC_TYPE_ROUTING_SLAVE)

class BASIC_TYPE_CONTROLLER(ZWClass):
    zid = 0x01
    name = 'BASIC_TYPE_CONTROLLER'
    short_name = 'controller'
    man_classes = []

ZWBasicDeviceClasses.register(BASIC_TYPE_CONTROLLER)



class BASIC_TYPE_STATIC_CONTROLLER(ZWClass):
    zid = 0x02
    name = 'BASIC_TYPE_STATIC_CONTROLLER'
    short_name='static_controller'
    man_classes = []

ZWBasicDeviceClasses.register(BASIC_TYPE_STATIC_CONTROLLER)

class GENERIC_TYPE_STATIC_CONTROLLER(ZWClass):
    zid = 0x02
    name = 'GENERIC_TYPE_STATIC_CONTROLLER'
    short_name='static_controller'
    man_classes =[]




class SPECIFIC_TYPE_PC_CONTROLLER(ZWClass):
    zid = 0x01
    name = 'SPECIFIC_TYPE_PC_CONTROLLER'
    short_name='pc_controller'
    man_classes = []
    
ZWGenericSpecificDeviceClasses.register(GENERIC_TYPE_STATIC_CONTROLLER, SPECIFIC_TYPE_PC_CONTROLLER)


class GENERIC_TYPE_GENERIC_CONTROLLER(ZWClass):
    zid = 0x01
    name = 'GENERIC_TYPE_GENERIC_CONTROLLER'
    short_name = 'generic_controller'
    man_classes = []

class SPECIFIC_TYPE_PORTABLE_REMOTE_CONTROLLER(ZWClass):
    zid = 0x01
    name = 'SPECIFIC_TYPE_PORTABLE_REMOTE_CONTROLLER'
    short_name = 'portable_remote_controller'
    man_classes = []

ZWGenericSpecificDeviceClasses.register(GENERIC_TYPE_GENERIC_CONTROLLER, SPECIFIC_TYPE_PORTABLE_REMOTE_CONTROLLER)

class GENERIC_TYPE_SWITCH_MULTILEVEL(ZWClass):
    zid = 0x11
    name = 'GENERIC_TYPE_SWITCH_MULTILEVEL'
    short_name = 'multilevel'
    man_classes = [0x20, 0x26]

class SPECIFIC_TYPE_POWER_SWITCH_MULTILEVEL(ZWClass):
    zid= 0x01
    name = 'SPECIFIC_TYPE_POWER_SWITCH_MULTILEVEL'
    short_name = 'power'
    man_classes = []


ZWGenericSpecificDeviceClasses.register(GENERIC_TYPE_SWITCH_MULTILEVEL, SPECIFIC_TYPE_POWER_SWITCH_MULTILEVEL)

ZWGenericSpecificDeviceClasses.register(GENERIC_TYPE_SWITCH_MULTILEVEL, SPECIFIC_TYPE_NONE)

class GENERIC_TYPE_SWITCH_BINARY(ZWClass):
    zid =0x10
    name ='GENERIC_TYPE_SWITCH_BINARY'
    short_name = 'binary'
    man_classes = [0x20, 0x25]

class SPECIFIC_TYPE_POWER_SWITCH_BINARY(ZWClass):
    zid= 0x01
    name = 'SPECIFIC_TYPE_POWER_SWITCH_BINARY'
    short_name = 'power'
    man_classes = []
    
ZWGenericSpecificDeviceClasses.register(GENERIC_TYPE_SWITCH_BINARY, SPECIFIC_TYPE_POWER_SWITCH_BINARY)

ZWGenericSpecificDeviceClasses.register(GENERIC_TYPE_SWITCH_BINARY, SPECIFIC_TYPE_NONE)




# Command Classes

class COMMAND_CLASS_BASIC(ZWClass):
    """
    
    #define BASIC_VERSION                                   0x01
    #define BASIC_SET                                       0x01
    #define BASIC_GET                                       0x02
    #define BASIC_REPORT                                    0x03
    """
    zid = 0x20
    name = 'COMMAND_CLASS_BASIC'
    short_name = 'basic'

    def __init__(self, holder):
        ZWClass.__init__(self, holder)
        self.reg(self.level_set)
        self.reg(self.level_get)
    
    
    def _limit_level(self, level):
        if level > 99:
            return 100
        else:
            return level

    def level_set(self, level):
        if level > 100:
            l = 0xff
        elif level == 100:
            l = 99
        else:
            l = level
        # TODO restrictions to valid level values
        self.holder.level = self._limit_level(int(l))
        self.holder.add_queue([self.zid, 0x01, int(l)])
        #self.holder.add_queue([self.zid, 0x02])
        
    def level_get(self):
        self.holder.send_app_command([self.zid, 0x02])
    
    def level_report(self):
        lev = self.holder.model.my_controller.level
        if lev > 99:
            l = 0xff;
        else:
            l = lev
        self.holder.send_app_command([self.zid, 0x03, l])
    
    def app_command_handler(self, command):
        cmd = command[1]
        if cmd == 0x03:
            self.holder.level = self._limit_level(command[2])
            logging.info('got level change to %d' % self._limit_level(command[2]))
        elif cmd == 0x02:
            #self.level_report()
            pass
        elif cmd == 0x01:
            self.holder.model.my_controller.level = self._limit_level(command[2])
    
    def unsol_node_info_handler(self, command):
        logging.info('checking level (basic) for node %d' % self.holder.node_id)
        self.level_get()
        pass
        
    def flash(self):
        """
        To emulate indicator behaviour, flash the device on and off with delays
        
        """
        orig_level = self.holder.level
        if self.holder.level < 0xff:
            new_level = 0xff
        else:
            new_level = 0
        self.holder.add_queue([self.zid, 0x01, new_level])
        #self.holder.add_queue([self.zid, 0x02])
        self.holder.add_queue([self.zid, 0x01, 0xff-new_level])
        #self.holder.add_queue([self.zid, 0x02])
        self.holder.add_queue([self.zid, 0x01, new_level])
        #self.holder.add_queue([self.zid, 0x02])
        self.holder.add_queue([self.zid, 0x01, orig_level])
        #self.holder.add_queue([self.zid, 0x02])



ZWCommandClasses.register(COMMAND_CLASS_BASIC)  

class COMMAND_CLASS_SWITCH_BINARY(ZWClass):
    zid=0x25
    name = 'COMMAND_CLASS_SWITCH_BINARY'
    short_name='switch_binary'

    def __init__(self, holder):
        ZWClass.__init__(self, holder)
        self.reg(self.level_set)
        self.reg(self.level_get)
    
    def level_set(self, level):
        # TODO restrictions to valid level values
        if level != 0:
            l = 0xff
        else:
            l = 0
        self.holder.level = self._limit_level(l)
        self.holder.add_queue([self.zid, 0x01, l])
        #self.holder.add_queue([self.zid, 0x02])
        
    def level_get(self):
        self.holder.send_app_command([self.zid, 0x02])
    
    def _limit_level(self, level):
        if level > 99:
            return 100
        else:
            return level
    
    def app_command_handler(self, command):
        cmd = command[1]
        if cmd == 0x03:
            self.holder.level = self._limit_level(command[2])
            logging.info('got level change to %d' % self._limit_level(command[2]))
    
    def unsol_node_info_handler(self, command):
        logging.debug("polling (binary)")
        self.level_get()
        pass
        
ZWCommandClasses.register(COMMAND_CLASS_SWITCH_BINARY)


class COMMAND_CLASS_SWITCH_MULTILEVEL(ZWClass):
    zid=0x26
    name = 'COMMAND_CLASS_SWITCH_MULTILEVEL'
    short_name='switch_multilevel'

    def __init__(self, holder):
        ZWClass.__init__(self, holder)
        self.reg(self.level_set)
        self.reg(self.level_get)

    def _limit_level(self, level):
        if level > 99:
            return 100
        else:
            return level

    
    def level_set(self, level):
        # TODO restrictions to valid level values
        l = int(level)
        if l > 99:
            l = 0xff
        self.holder.level = self._limit_level(int(l))
        logging.debug('level_set: holder nodeid is %d' % int(self.holder.node_id))
        self.holder.add_queue([self.zid, 0x01, int(l)])
        #self.holder.add_queue([self.zid, 0x02])
        
    def level_get(self):
        self.holder.send_app_command([self.zid, 0x02])
    
    def app_command_handler(self, command):
        cmd = command[1]
        if cmd == 0x03:
            self.holder.level = self._limit_level(command[2])
            logging.info('got level change to %d' % self._limit_level(command[2]))
    
    def unsol_node_info_handler(self, command):
        logging.debug("polling (multilevel)")
        self.level_get()
        pass
        
ZWCommandClasses.register(COMMAND_CLASS_SWITCH_MULTILEVEL)
    
class COMMAND_CLASS_SWITCH_ALL(ZWClass):
    zid=0x27
    name ='COMMAND_CLASS_SWITCH_ALL'
    short_name='switch_all'

ZWCommandClasses.register(COMMAND_CLASS_SWITCH_ALL)

class COMMAND_CLASS_MANUFACTURER_SPECIFIC(ZWClass):
    """
    #define MANUFACTURER_SPECIFIC_VERSION                   0x01
    #define MANUFACTURER_SPECIFIC_GET                       0x04
    #define MANUFACTURER_SPECIFIC_REPORT                    0x05
    """
    
    zid=0x72
    name = 'COMMAND_CLASS_MANUFACTURER_SPECIFIC'
    short_name = 'manufacturer_specific'
    
    def __init__(self, holder):
        ZWClass.__init__(self, holder)
        self.reg(self.ms_get)
        self.reg(self.ms_report)

    def ms_get(self):
        """ 
        issue get version for this device.. to be filled in by version_report incoming
        """
        self.holder.send_app_command([self.zid, 0x04])

    def ms_report(self, command):
        """ report our version info -- if we are a controller, no otherwise"""
 
        myms = self.holder.model.my_controller.manufacturer_specific.get_ms()
        self.holder.send_app_command([self.zid, 0x05] + myms)

    def app_command_handler(self, command):
        """ 
        respond as we need to..
        """
        cmd = command[1]
        if cmd == 0x04:
            # handle version get
            # issue a version report
            self.ms_report(command)
        elif cmd == 0x05:
            # save report info
            ms = ManufacturerSpecific()
            ms.set_ms_from_report(command[2:])
            self.holder.manufacturer_specific = ms
    
    def unsol_node_info_handler(self, command):
        """request manuf info if we don't already have it"""
        if self.holder.manufacturer_specific.is_null():
            self.ms_get()


ZWCommandClasses.register(COMMAND_CLASS_MANUFACTURER_SPECIFIC)

class COMMAND_CLASS_PROTECTION(ZWClass):
    zid = 0x75
    name = 'COMMAND_CLASS_PROTECTION'
    short_name = 'protection'

ZWCommandClasses.register(COMMAND_CLASS_PROTECTION)

class COMMAND_CLASS_VERSION(ZWClass):
    """
    #define COMMAND_CLASS_VERSION                           0x86
    #define VERSION_VERSION                                 0x01
    #define VERSION_GET                                     0x11
    #define VERSION_REPORT                                  0x12
    #define VERSION_COMMAND_CLASS_GET                       0x13
    #define VERSION_COMMAND_CLASS_REPORT                    0x14
    
    """
    zid = 0x86
    name = 'COMMAND_CLASS_VERSION'
    short_name = 'version'
    
    def __init__(self, holder):
        ZWClass.__init__(self, holder)
        self.reg(self.version_get)
        self.reg(self.version_report)
        self.reg(self.version_command_class_get)
        self.reg(self.version_command_class_report)

    def version_get(self):
        self.holder.send_app_command([self.zid, 0x11])
        
    def version_report(self):
        """report on our version, only if we are the controller, otherwise nope"""
        # gather information
        lib_type = 0x01
        zwave_protocol_vers = 2
        zwave_protocol_sub_vers = 17
        app_vers = 1
        app_sub_vers = 0
        self.holder.send_app_command([self.zid, 0x12, lib_type, zwave_protocol_vers, zwave_protocol_sub_vers, app_vers,  app_sub_vers])
        
    def version_command_class_get(self, command_class_no):
        self.holder.send_app_command([self.zid, 0x13, command_class_no])
    
    def version_command_class_report(self, command_class_no):
        # find command class version
        cc = self.holder.model.my_controller.comclass
        if cc.has_key(command_class_no):
            ccvers = cc[command_class_no].get_version()
            logging.debug('command class %s has version %s' % (hex(command_class_no), hex(ccvers)))
        else:
            logging.debug('command class %s not supported, so version = 1' % hex(command_class_no))
            ccvers = 1
        self.holder.send_app_command([self.zid, 0x14, command_class_no, ccvers])

    def app_command_handler(self, command):
        cmd = command[1]
        if cmd == 0x11:
            # handle version get
            # issue a version report
            self.version_report()
        elif cmd == 0x12:
            # save the report
            self.holder.version = command[2:7]
        elif cmd == 0x13:
            self.version_command_class_report(command[2])
        elif cmd == 0x14:
            self.holder.ccversions[command[2]] = command[3]
        else:
            pass
            
ZWCommandClasses.register(COMMAND_CLASS_VERSION)




class COMMAND_CLASS_ALARM(ZWClass):
    """
    /* Alarm command class commands */
    #define ALARM_VERSION                                   0x01
    #define ALARM_GET                                       0x04
    #define ALARM_REPORT                                    0x05
    """
    
    zid=0x71
    name='COMMAND_CLASS_ALARM'
    short_name='alarm'

    def __init__(self, holder):
        ZWClass.__init__(self, holder)
        # initialize the holder object
        self.holder.alarms={}
        self.holder.alarms[0] = 0
        self.holder.alarms[1] = 0
        self.reg(self.alarm_get)

    def alarm_get(self, alarm_type):
        self.holder.send_app_command([self.zid, 0x04, alarm_type])
    
    def app_command_handler(self, command):
        cmd = command[1]
        if cmd == 0x05: #alarm report
            alarm_type = command[2]
            alarm_level = command[3]
            logging.info('got alarm report for %d is %d' % (alarm_type, alarm_level))
            self.holder.set_alarm(alarm_type, alarm_level)


    def unsol_node_info_handler(self, command):
        self.alarm_check(command)
        
    def alarm_check(self, command):
        """
        check set of alarms for this device.  We expect 0 and 1 to be defined
        BUT do we need a manufacturer check here?
        """
        logging.debug('checking alarms for node %d' % self.holder.node_id)
        self.alarm_get(0)
        self.alarm_get(1)
        

ZWCommandClasses.register(COMMAND_CLASS_ALARM)

class COMMAND_CLASS_ASSOCIATION(ZWClass):
    """
    #define ASSOCIATION_VERSION                             0x01
    #define ASSOCIATION_VERSION_V2                          0x02
    #define ASSOCIATION_SET                                 0x01
    #define ASSOCIATION_GET                                 0x02
    #define ASSOCIATION_REPORT                              0x03
    #define ASSOCIATION_REMOVE                              0x04
    #define ASSOCIATION_GROUPINGS_GET                       0x05
    #define ASSOCIATION_GROUPINGS_REPORT                    0x06
    #define ASSOCIATION_COMPOSITE_CMD_ADD_V2                0x07
    #define ASSOCIATION_COMPOSITE_CMD_REMOVE_V2             0x08
    #define ASSOCIATION_COMPOSITE_CMD_GET_V2                0x09
    #define ASSOCIATION_COMPOSITE_CMD_REPORT_V2             0x0A
    """
    zid=0x85
    name='COMMAND_CLASS_ASSOCIATION'
    short_name='association'
    
    def __init__(self, holder):
        ZWClass.__init__(self, holder)
    
    def app_command_handler(self, command):
        """ 
        respond as we need to..
        """
        cmd = command[1]
        #TODO, dummy responses
        logging.info('got association cmd %d' % cmd)

ZWCommandClasses.register(COMMAND_CLASS_ASSOCIATION)




class COMMAND_CLASS_INDICATOR(ZWClass):
    """Indicator command class

    #define INDICATOR_VERSION                               0x01
    #define INDICATOR_SET                                   0x01
    #define INDICATOR_GET                                   0x02
    #define INDICATOR_REPORT                                0x03
    
    """
    zid = 0x87
    name = 'COMMAND_CLASS_INDICATOR'
    short_name = 'indicator'

    def __init__(self, holder):
        ZWClass.__init__(self, holder)
        self.reg(self.level_set)
        self.reg(self.level_get)
        self.reg(self.flash)
    
    def level_set(self, level):
        if level != 0:
            level = 0xff
        # TODO restrictions to valid level values
        self.holder.level = int(level)
        self.holder.add_queue([self.zid, 0x01, int(level)])
        self.holder.add_queue([self.zid, 0x02])
        
    def level_get(self):
        self.holder.send_app_command([self.zid, 0x02])
    
    def app_command_handler(self, command):
        cmd = command[1]
        if cmd == 0x03:
            self.holder.level = command[2]
            logging.info('got level change to %d' % command[2])
    
    def unsol_node_info_handler(self, command):
        logging.info('checking level for node %d' % self.holder.node_id)
        self.level_get()
    
    def flash(self):
        """
        To emulate indicator behaviour, flash the device on and off with delays
        
        """
        orig_level = self.holder.level
        if self.holder.level < 0xff:
            new_level = 0xff
        else:
            new_level = 0
        self.holder.add_queue([self.zid, 0x01, new_level])
        self.holder.add_queue([self.zid, 0x02])
        self.holder.add_queue([self.zid, 0x01, 0xff-new_level])
        self.holder.add_queue([self.zid, 0x02])
        self.holder.add_queue([self.zid, 0x01, new_level])
        self.holder.add_queue([self.zid, 0x02])
        self.holder.add_queue([self.zid, 0x01, orig_level])
        self.holder.add_queue([self.zid, 0x02])

ZWCommandClasses.register(COMMAND_CLASS_INDICATOR)

class  COMMAND_CLASS_CONTROLLER_REPLICATION(ZWClass):
    """Controller replication

    #define COMMAND_CLASS_CONTROLLER_REPLICATION            0x21
    /* Controller Replication command class commands */
    #define CTRL_REPLICATION_TRANSFER_VERSION               0x01
    #define CTRL_REPLICATION_TRANSFER_GROUP                 0x31
    #define CTRL_REPLICATION_TRANSFER_GROUP_NAME            0x32
    #define CTRL_REPLICATION_TRANSFER_SCENE                 0x33
    #define CTRL_REPLICATION_TRANSFER_SCENE_NAME            0x34
    /* Max. group and scene text length */
    #define CTRL_REPLICATION_NAME_MAX_TXT_LEN               0x0C

    """
    zid = 0x21
    name = 'COMMAND_CLASS_CONTROLLER_REPLICATION'
    short_name = 'replication'

    def __init__(self, holder):
        ZWClass.__init__(self, holder)

    def send_done(self):
        rcmd = ZWReplicationReceiveComplete(self.holder.model.network.serial_api)
        rcmd.go()

    def app_command_handler(self, command):
        cmd = command[1]
        if cmd >= 0x31 and cmd <= 0x34:
            logging.info('got replicate send cmd %d' % cmd)
            self.send_done()

ZWCommandClasses.register(COMMAND_CLASS_CONTROLLER_REPLICATION)

if __name__ == '__main__':
    unittest.main()