#!/usr/bin/env python
# encoding: utf-8
"""
zwclassresolver.py

Defining the base classses for behaviours and then their command sets and attributes

Created by Graeme Sutherland on 2008-04-25.
Copyright (c) 2008 TrickleStar Limited. All rights reserved.
"""

import sys
import os
import unittest
import exceptions
import inspect

from zwclassresolver import *
from zwtransport import *

class ZWBasicDeviceClasses(object):
    """ Store/selector for the basic device classes """
    classes = {}
    
    #@classmethod
    def register(cls, cmd_class):
        cls.classes[cmd_class.zid] = cmd_class
    register = classmethod(register)

    #@classmethod
    def get_class(cls, zid):
        if cls.classes.has_key(zid):
            return cls.classes[zid]
        else:
            logging.warning("Unknown basic device class for zid=" + str(zid))
            return None
    get_class = classmethod(get_class)


class ZWGenericSpecificDeviceClasses(object):
    """ the ID need to be a tuple when getting classes and when registering """
    
    classes = {}
    
    #@classmethod
    def register(cls, gen, spec):
        cls.classes[(gen.zid, spec.zid)] = (gen, spec)
    register = classmethod(register)

    #@classmethod
    def get_class(cls, gen_zid, spec_zid):
        if cls.classes.has_key((gen_zid, spec_zid)):
            return cls.classes[(gen_zid, spec_zid)]
        elif cls.classes.has_key((gen_zid, 0x00)):
            return cls.classes[(gen_zid, 0x00)]            
        else:
            return (GENERIC_TYPE_NONE, SPECIFIC_TYPE_NONE)
    get_class = classmethod(get_class)
    
class ZWCommandClasses(object):
    """ the set of command classes """
    classes = {}
    
    #@classmethod
    def register(cls, cmd_class):
        cls.classes[cmd_class.zid] = cmd_class
    register = classmethod(register)

    #python 2.3 sigh
    #@classmethod
    def get_class(cls, zid):
        if cls.classes.has_key(zid):
            return cls.classes[zid]
        else:
            return COMMAND_CLASS_NOT_IMPLEMENTED
    get_class = classmethod(get_class)

class ZWClass(object):
    """
    Defines Z-Wave command  or device class
    """
    zid = 0 # override in subclasses
    short_name = "commandclassunrefined"
    name="Command Class (unrefined)"

    def __init__(self, holder):
        self.holder = holder
        self.cmds={}
        self.cc_version=1

    def run(self, command, *args, **kw):
        """
        Run this command. So, resolve to a method and pass args
        """
        if self.cmds.has_key(command):
            return self.cmds[command](*args, **kw)
        else:
            raise ZWClassNoCommandError( "no command found named " + command )

    def reg(self, f):
        self.cmds[f.__name__] = f
    
    def app_command_handler(self, command):
        pass
    
    def unsol_node_info_handler(self, command):
        """
        Do a poll or something when we get an unsolicited node info received
        """
        pass

    def get_version(self):
        """
        Return the command class version
        
        """
        return self.cc_version

class SPECIFIC_TYPE_NONE(ZWClass):
    zid= 0x00
    name = 'SPECIFIC_TYPE_NONE'
    short_name = 'specific_none'
    man_classes = []

class GENERIC_TYPE_NONE(ZWClass):
    zid=0
    name='GENERIC_TYPE_NONE'
    short_name = 'none'
    man_classes =[]

class COMMAND_CLASS_NOT_IMPLEMENTED(ZWClass):
    zid=-1
    name= 'COMMAND_CLASS_NOT_IMPLEMENTED'
    short_name = 'not_implemented'

class ZWClassNoCommandError(exceptions.Exception):
    """no command available"""
    def __init__(self, args=None):
        args = args

if __name__ == '__main__':
    unittest.main()