#!/usr/bin/env python
# encoding: utf-8
"""
zwserver.py

Simple web server to access Z-Wave

Created by Graeme Sutherland on 2008-05-06.
Copyright (c) 2008 TrickleStar Limited. All rights reserved.
"""

import sys
import os
import time
import BaseHTTPServer
import unittest
import logging
import simplejson
import sys
import logging
import threading
import signal
import urllib
import random
from optparse import OptionParser


#Platform-based messing around
plat='x86'
osv = '10.5'
vi = sys.version_info
if vi[1] == 3:
    osv='10.4'
if sys.byteorder=='big':
    plat='ppc'

slibpath="./libs/%s/%s" % (plat, osv)
sys.path = [slibpath, './growl', './sleepcatch'] + sys.path

from _sleepcatch import *

from serial import *
from zwserialapi import *
from zwbasecommand import *
from zwcontroller import *
from zwtransport import *
from zwmodel import *
from zwclassresolver import *
from zwclasses import *
from zwconfig import *
from notify import Notifier
from sleepcatch import Sleepcatch

class ZWServer(object):

    serial= None

    def __init__(self):
        pass

    def run(self, options, args, my_thread):
        self.my_thread = my_thread
        self.bind_server(options, args)
        self.start_server(options, args)

    def bind_server(self, options, args):
        #logging.info('binding server')
        ZWHandler.my_thread = self.my_thread
        try:
            self.server = BaseHTTPServer.HTTPServer((options.host, options.port), ZWHandler)
            return True
        except Exception, e:
            #logging.warning('Cant bind server to address')
            time.sleep(2)
            # look out.. kill other server and then we can start
            try:
                #logging.warning('stopping existing server')
                r = urllib.urlopen('http://127.0.0.1:4344/ezw/sys/stop')
                r.close()
                time.sleep(1)
            except Exception, ee:
                #logging.warning('can not kill existing server')
                try:
                    r.close()
                except:
                    pass
                time.sleep(1)
                return False
            return False

    def start_server(self, options, args):
        """ Start the server on the give port, ought to 
        give us an exception if already running, so that helps (for now)"""
        #logging.info('starting server')
        if self.server:
            self.server.serve_forever()
    
    def stop(self):
        try:
            self.server.server_close()
        except:
            pass 


class ZWHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    
    my_thread = None
    
    def log_message(self, format, *args):
        # disable http-level logging for now
        pass
    
    def get_model(self):
        if self.my_thread:
            return self.my_thread.get_model()
        else:
            return None
        
    def get_log(self):
        if self.my_thread:
            return self.my_thread.get_log()
            
    def do_GET(self):
        try:
            self.run_command()
        except:
            logging.exception('exception during GET')
    
    def do_POST(self):
        try:
            logging.debug('got POST')
            data_size = self.headers.getheader('Content-Length')
            data = self.rfile.read(int(data_size))
            logging.debug('data is ' + data)
            obj = simplejson.loads(data)
            if obj and obj.has_key('subject') and obj.has_key('verb'):
                # yes
                sub = obj['subject']
                if sub == 'sys':
                    r = self.sys(obj)
                elif sub == 'net':
                    r = self.net(obj)
                elif sub == 'dev':
                    r = self.dev(obj)
                else:
                    r = {}
            else:
                self.send_error(404, 'missing subject and or verb');
        except:
            logging.exception('exception during post')

    def run_command(self):
        """docstring for run_command"""
        args = self.path.split('/')
        # print args
        if len(args) > 2:
            if args[1] != 'ezw':
                self.send_error(404 , "Bad Selector")
            else:
                subject = args[2]
                self.despatch(subject, args[2:])
        else:
            self.send_error(404, 'Missing selector and subject and args')
    
    def despatch(self, subject, args):
        """docstring for depatch"""
        logging.debug('subject is ' + subject)
        if subject == 'sys':
            self.getsys(args)
        elif subject == 'net':
            self.net(args)
        elif subject == 'dev':
            self.dev(args)
        else:
            self.send_error(404, 'Unknown subject')
    
    def sys(self, args):
        verb = args['verb']
        if verb == 'check':
            logging.debug("(check) Widget checked server")
            self.ok("All good")
        if verb == 'status':
            # return status info
            model = self.get_model()
            if model:
                self.json_ok(model.get_status())
            else:
                self.json_ok({'serial': 'unknown', 'model': 'unknown'})
        elif verb == 'stop':
            self.ok("Stopping")
            logging.info('(stop) Stopping')
            sys.exit(0)
        elif verb == 'set_config':
            # save configuration
            put_config(args['config'])
            logging.debug('config changed')
            log_config()
            save_config(self.get_model())
            self.json_ok([True])
        else:
            self.send_error(404, 'Unknown sys verb')

    def getsys(self, args):
        if args[1] == 'stop':
            self.ok("Stopping")
            logging.info('(stop) Stopping')
            self.my_thread.exit()

    def net(self, args):
        verb = args['verb']
        logging.debug('verb is ' + verb)
        if verb == 'nodelist':
            model = self.get_model()
            if model:
                self.json_ok(model.get_nodelist())
            else:
                self.json_ok([])
        elif verb == 'update':
            model = self.get_model()
            if model:
                model.update_ping()
                response = {}
                devs = []
                for n in model.devices.keys():
                    devs.append(model.devices[n].get_info())
                model.mutiply_devices(devs)
                response['devices'] = devs
                response['zwave'] = (len(devs) > 1) 
                response['usb'] = model.is_usb_ok()
                response['mode'] = model.mode
                response['client_command'] = model.client_command.get_command()
                self.json_ok(response)
            else:
                self.json_ok([])
        elif verb == 'clear_client_command':
            model = self.get_model()
            if model:
                model.client_command.done_command()
                self.json_ok([])                
        elif verb == 'start_remove':
            model = self.get_model()
            if model:
                self.json_ok(model.start_remove())
            else:
                self.json_ok([])
        elif verb == 'stop_remove':
            model = self.get_model()
            if model:
                self.json_ok(model.start_remove())
            else:
                self.json_ok([])
        elif verb == 'start_add':
            model = self.get_model()
            if model:
                self.json_ok(model.start_add(args))
            else:
                self.json_ok([])
        elif verb == 'stop_add':
            model = self.get_model()
            if model:
                self.json_ok(model.stop_add())
            else:
                self.json_ok([])
        elif verb == 'reset_controller':
            model = self.get_model()
            if model:
                self.json_ok(model.reset_controller())
            else:
                self.json_ok([])
        elif verb == 'system_sleep':
            model = self.get_model()
            if model:
                self.json_ok(model.system_sleep())
            else:
                self.json_ok({})
        elif verb == 'system_startup':
            model = self.get_model()
            if model:
                self.json_ok(model.system_startup())
            else:
                self.json_ok({})
        elif verb == 'system_shutdown':
            model = self.get_model()
            if model:
                self.json_ok(model.system_shutdown())
            else:
                self.json_ok({})
        elif verb == 'add_to_network':
            model = self.get_model()
            if model:
                self.json_ok(model.add_to_network())
            else:
                self.json_ok({})
        elif verb == 'request_network_update':
            model = self.get_model()
            if model:
                self.json_ok(model.request_network_update())
            else:
                self.json_ok({})
        elif verb == 'get_net_log':
            log = self.get_log()
            if log:
                self.json_ok(log.get_log())
            else:
                self.json_ok({})
        elif verb == 'clear_net_log':
            log = self.get_log()
            if log:
                log.clear()
                self.json_ok(log.get_log())
            else:
                self.json_ok({})                               
        else:
            self.send_error(404, 'Unknown verb')

    def dev(self, args):
        nodeid = args['nodeid']
        model = self.get_model()
        if model and model.devices.has_key(nodeid):
            self.json_ok(model.devices[nodeid].go(args));
        else:
            self.json_ok([])
    
    def ok(self, result_body):
        self.send_response(200)
        self.send_header('Status', '200 OK')
        self.send_header('Content-type', 'text/plain')
        self.end_headers()
        self.wfile.write(result_body)
    
    def json_ok(self, obj):
        self.send_response(200)
        self.send_header('Status', '200 OK')
        self.send_header('Content-type', 'text/plain')
        self.end_headers()
        self.wfile.write(simplejson.dumps(obj))
        
def get_options(argv):
    parser = OptionParser("usage: %prog [options]")
    parser.add_option("-q", "--quiet",
                      action="store_false", dest="verbose", default=True,
                      help="don't print status messages to stdout")
    parser.add_option("-d", "--device", action="store", type="string", dest="device", default="/dev/cu.usbserial")
    parser.add_option("-p", "--port", action="store", type="int", dest="port", default=4344)
    parser.add_option("-b", "--bind", action="store", type="string", dest="host", default="localhost")
    parser.add_option("-f", "--fast", action="store_true", dest="fast", default=False)
    parser.add_option("-l", "--logtofile", action="store_true", dest="file_log", default=False)
    parser.add_option("-F", "--randomfail", action="store_true", dest="random_fail", default=False)
    parser.add_option("-m", "--multiply", action="store", type="int", dest="multiply", default=-1)
    
    return parser.parse_args()

def start_logging(options, args):
    if options.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO
    #logging.basicConfig()
    logger = logging.getLogger()
    logger.setLevel(level)
    if options.file_log:
        f = logging.FileHandler('/tmp/zwserver.log','w' )
    else:
        f = logging.StreamHandler();
    f.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s'))
    logger.addHandler(f)



class ServerThread(threading.Thread):
    
    quit = False
    
    def __init__(self, options, args, model_thread):
        self.options = options
        self.args = args
        self.model_thread = model_thread
        self.quit = False
        threading.Thread.__init__(self)
        
    def run(self):
        logging.info('starting server thread')
        while not self.quit:
            try:
                self.server = ZWServer()
                self.server.run(self.options, self.args, self)
            except:
                logging.exception('server exception, restarting')
            if self.server:
                self.server.stop()
    
    def get_model(self):
        if self.model_thread:
            return self.model_thread.get_model()
        else:
            return None
    
    def get_log(self):
        if self.model_thread:
            return self.model_thread.log
    
    def exit(self):
        if self.model_thread:
            self.model_thread.quit = True
        self.quit = True
            
   
class ModelThread(threading.Thread):
    
    model = None
    zws = None
    quit = False

    def __init__(self, options, args):
        self.options = options
        self.args = args
        self.log = DeviceLog(50)
        threading.Thread.__init__(self)

    def run(self):
        logging.info('starting model thread')
        while not self.quit:
            try:
                self.start_serial(self.options, self.args)
                if self.zws.serial:
                    self.start_model(self.options, self.args)
                    self.run_model()
            except Exception, e:
                logging.exception('model error, restarting')
                self.log.error('model error, restarting')
            self.stop_model()
            self.stop_serial()
            time.sleep(2)
            
    
    def start_serial(self, options, args):
        self.zws = ZWSerialAPI()
        self.zws.set_device(options.device)
        self.zws.start()

    def start_model(self, options, args):
        m = self.model
        self.model = ZWModel(self.log)
        self.model.set_multiply(options.multiply)
        self.model.set_serial(self.zws)
        if m == None or m.startup == True:
            self.model.set_startup()
        self.model.build()
    
    def run_model(self):
        while not self.quit:
            # loop processing
            if self.options.random_fail and random.random() > 0.01:
                raise Exception("Random failure (from -F flag)")
            self.model.poll()
    
    def stop_model(self):
        self.model = None
        
    def get_model(self):
        return self.model
    
    def stop_serial(self):
        try:
            if self.zws:
                self.zws.stop()
            self.zws = None
        except:
            pass
            
    
    def signal_quit(self):
        self.quit = True





class PowerThread(threading.Thread):
    def __init__(self, m):
        self.m = m
        threading.Thread.__init__(self)
        
    def run(self):
        self.sc = Sleepcatch(self.powerinfo)
        while True:
            poll_loop()
            time.sleep(0.1)
    
    def powerinfo(self, event):
        logging.debug('System Power event: %s' % event)
        if event == 'can_system_sleep':
            model = m.get_model()
            if model:
                model.system_sleep()
                time.sleep(5)
        elif event == 'system_will_sleep':
            model = m.get_model()
            if model:
                model.system_sleep() 
                time.sleep(5)       
        elif event == 'system_powered_on':
            model = m.get_model()
            if model:
                model.set_wake_from_sleep()                    
        else:
            pass
        

m = None
s = None

def start_threads(options, args):
    global m
    global s
    load_config()
    m = ModelThread(options, args)
    m.start()
    s = ServerThread(options, args, m)
    s.start()
    p = PowerThread(m)
    p.start()
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, sigterm_handler)
    while (not m.quit) and (not s.quit):
        time.sleep(1)
    

def sigterm_handler(signum, frame):
    global m
    logging.info('SIGTERM captured, running shutdown')
    if m and m.model:
        m.model.system_shutdown()
    time.sleep(5)
    sys.exit(0)

    
def main():
    (options, args) = get_options(sys.argv)
    start_logging(options, args)
    start_threads(options, args)


        
if __name__ == '__main__':
    sys.exit(main())
    
    