#!/usr/bin/env python
# encoding: utf-8
"""
zwtransport.py

Created by Graeme Sutherland on 2008-05-01.
Copyright (c) 2008 TrickleStar Limited. All rights reserved.
"""

import sys
import os
import unittest
import time

from zwconstants import *
from zwbasecommand import *
from zwserialapi import *


class ZWSendData(ZWBaseCommand):
    """
    HOST->ZW: REQ | 0x13 | nodeID | dataLength | pData[] | txOptions | funcID 
    0x1,0x9,0x0,0x13,0x2,0x2,0x1,0xff,0x4,0xf1,0xee]

    """
    def __init__(self, serial_api, dest_node, data, tx_options = (TRANSMIT_OPTION_AUTO_ROUTE | TRANSMIT_OPTION_ACK) ):
        ZWBaseCommand.__init__(self, serial_api)
        self.name='ZWSendData'
        self.wait_for_response = True
        self.frame = ZWFrame(REQ)
        self.frame.set_command([0x13, dest_node, len(data) ] + data + [tx_options, fid.get_fid(dest_node)])
        self.match=[[RES,0x13],[REQ,0x13]]

    def go(self, mydev):
        self.send()
        if len(self.match[1]) == 3:
            rframe = self.match[1][2]
            tx_status = rframe.command[2]
            if tx_status == TRANSMIT_COMPLETE_OK:
                # successful transmit
                mydev.tx_confirm()
            elif tx_status == TRANSMIT_COMPLETE_NO_ACK:
                # remote unit did not respond
                mydev.tx_no_ack()
            elif tx_status == TRANSMIT_COMPLETE_FAIL:
                # soft failure
                mydev.model.log.error('network transmission failure')
                mydev.tx_no_ack()
            else:
                pass

class FID(object):
    """
    An instance of a func id
    """
    def __init__(self, id, node_id):
        self.fid = fid
        self.node_id = node_id
        self.sendtime = time.time()


class FuncIdManager(object):
    """
    Manage func ids for commands. This is principally so we can allocate funcids and associate
    node_ids with them so we can find out who is failing to respond etc
    
    """
    def __init__(self):
        self.max_fid = 0xff
        self.next_fid = 0x01
        self.fids= {}
    
    def get_fid(self, node_id):
        fidid = self.next_fid
        self.next_fid = (self.next_fid + 1) % self.max_fid
        fid = FID(fidid, node_id)
        self.fids[fidid] = fid
        return fidid
    
    def lookup_fid(self, fid):
        return self.fids[fid]

fid = FuncIdManager()


    

if __name__ == '__main__':
    for f in range(0x102):
        ff = fid.get_fid(1)
        print ff
        print fid.lookup_fid(ff).sendtime
        