/* 
 * Copyright 2008 TrickleStar Ltd
 * All Rights Reserved.
 * http://tricklestar.com
 *
 */


//
// base class for handling Z-Wave interface
// -- support 1 or more devices 

function ZWave() {

    this.devices = new Array();
    this.devSorted = new Array();
    this.USBState = false;
    this.ZWaveState = false;
    this.enableDeviceTable = false;
    this.widgetMode = 'run';
    this.myController = null;
    this.log = null;
    this.sortPlan = null;
    
    this.addDevice = function(dev) {
        this.devices.push(dev);
    };
    
    this.getDevice = function(index) {
        return this.devices[index];
    }; 
    
    this.getMyController = function() {
        for (i in this.devices) {
            if (this.devices[i].isMyController()) {
                return this.devices[i];
            }
        }
        return null;
    }
    
    this.getFrontDevice = function(index) {
        var count = 0;
        for (i in this.devSorted) {
            if (this.devSorted[i].front && !this.devSorted[i].isMyController()) {
                count++;
            }
            if (count == index) {
                return this.devSorted[i];
            }
        }
        return null;
    };
    
    this.getDeviceIndex = function(nodeid) {
        var count = 0;
        for (i in this.devSorted) {
            if (this.devSorted[i].front && !this.devSorted[i].isMyController()) {
                count++;
            }
            if (nodeid == this.devSorted[i].node_id) {
                return count;
            }
        }
        return -1;
    };
    
    this.getUSBInterfaceState = function() {
        return this.USBState;
    };
    
    this.setServerStatus = function(statinfo) {
    };
    
    this.getZWaveState = function() {
        return this.ZWaveState;
    };
    
    this.renderDevicesToFrontSide = function() {
        var divPrefix = "box";
        var dispPrefix = 'disp-dev-';
        var index=1;
        var count=0;
        for (idx in this.devSorted) {
            var ni = this.devSorted[idx].node_id;
            this.devSorted[idx].front = (widget.preferenceForKey("dev." + ni + ".front") == 'yes');
            if (this.devSorted[idx].front && !this.devSorted[idx].isMyController()) {
                var en = divPrefix + index;
                var el = document.getElementById(en);
                var dn = document.getElementById(dispPrefix + index);
                if (el && dn) {
                    this.devSorted[idx].renderUpdate(el, index);
                    el.style.display='block';
                    dn.style.display='block';
                    index += 1;
                }
                count += 1;
            }
        }
        for (var i=count+1;i<=8;i++) {
            var en = divPrefix + i;
            var el = document.getElementById(en);
            if (el) {
                el.style.display='none';
                el.innerHTML="";
            }
            var disp = document.getElementById(dispPrefix + i);
            if (disp) {
                disp.style.display = 'none';
                    }        
        }
        var scroll = document.getElementById('scrollAreaFront');
        scroll.object.refresh();
    };
    
    this.clearFrontSide = function() {
        for (var i=1;i<=8;i++) {
            var en = 'box' + i;
            var el = document.getElementById(en);
            if (el) {
                el.innerHTML="";
                el.style.display = 'none';
            }
            var ed = document.getElementById('disp-dev-' + i);
            if (ed) {
                ed.style.display = 'none';
            }
        }
        for (d in this.devices) {
            this.devices[d].slider = null;
        }
    };

    
    this.zwstart = function() {
        this.updateState();
    };
    
    this.clearClientCommand = function() {
        serverSend('net', 'clear_client_command', true, null);
    };
    
    this.delConfig = function(node) {
      var dev = this.getDeviceByNode(node);
      setPreferenceForKey(null, 'dev.' + node + '.front');
      setPreferenceForKey(null, 'dev.' + node + '.name');
      setPreferenceForKey(null, 'dev.' + node + '.config');
      if (dev) {
        this.clearClientCommand();
        dev.newConfig();
        dev.saveConfig();
      }
    };
    
    this.updateState = function() {
        var update = serverSend('net', 'update', false, null);
        if (update) {
            if (this.widgetMode == 'include' || this.widgetMode == 'exclude') {
                if (update.mode == 'run') {
                    var b = document.getElementById('dialog-box');
                    b.style.display='none';
                    this.widgetMode='run';
                }
            }
            else if (update.mode == 'include') {
                this.widgetMode = 'include';
            }
            else if (update.mode == 'exclude') {
                this.widgetMode = 'exclude';
            }
            var c = update.client_command;
            if (c) {
                 if (c.cmd == 'del_config') {
                    var no = c.node;
                    this.delConfig(no);
                    this.clearClientCommand();
                }
                else if (c.cmd == 'show_error') {
                    var no = c.node;
                    this.showErrorForNode(no);
                    this.clearClientCommand();                
                }
            }
            
            var now = (new Date()).getTime();
            this.USBState = update.usb;
            this.ZWaveState = update.zwave;
            for (var un in update.devices) {
                var up = update.devices[un];
                if (up != null) {
                    // look for existing node
                    var ni = up.nodeid;
                    var dev = this.getDeviceByNode(ni);
                    var dt = this.getDevtype(up.generic);
                    if (up.controller_capabilities) {
                        dt = 'USB Stick';
                    }
                    if (dev) {
                        // existing, so update
                        dev.deviceType = dt;
                        //dev.name = dt;
                        dev.updateLevelFromZWave(up.level);
                        dev.prod_type = up.generic;
                        dev.lib_type = up.base;
                        dev.lib_version = up.specific;
                        dev.last_update = now;
                        dev.controller_capabilities = up.controller_capabilities;
                        dev.details = up.details;
                        dev.comm_failure = up.comm_failure;
                        dev.alarms = up.alarms;
                        dev.home_id = up.homeid;
                        if (dev.controller_capabilities) {
                            this.myController = dev;
                        }
                    }
                    else {
                        // new, so create
                        this.addDevice(new Device(dt, dt, 1, false, up.level, 0, 0, 0x006b, up.generic, up.base, up.specific, 1, 44, ni));
                        var d = this.getDeviceByNode(ni);
                        d.getConfig();
                        d.saveConfig(); // and send to server
                        d.last_update = now;
                        d.controller_capabilities = up.controller_capabilities;
                        d.details = up.details;
                        d.comm_failure = up.comm_failure;
                        d.alarms = up.alarms;
                        d.home_id = up.homeid;
                        if (d.controller_capabilities) {
                            this.myController = d;
                        }
                    }
                }
                
            } // for
            // delete things not updated
            var delflag = false;
            for (var dd in this.devices) {
              var testdev = this.devices[dd];
                if (testdev.last_update < now) {
                    //this.device.splice(dd,1);
                    delete this.devices[dd];
                    delflag=true;
                }
            }
            if (delflag) {
                var newdev = Array();
                var newindex = 0;
                for (var dd in this.devices) {
                    if (this.devices[dd] != undefined) {
                        newdev[newindex] = this.devices[dd];
                        newindex = newindex + 1; 
                    }
                }
                this.devices = newdev;
            }
            this.doSort();
            renderStatus();
            renderUSBStickStatus();
            this.renderDevicesToFrontSide();
            if (this.enableDeviceTable) {
                this.renderDeviceTable(document.getElementById('box-content'));
            }
        } // if update
        else {
            this.USBState = false;
            this.ZWaveState = false;

        }
    };


    this.getDevtype = function(gentype) {
        var devtype = 'Relay';    
        if (gentype == 'GENERIC_TYPE_SWITCH_MULTILEVEL') {
            devtype = 'Dimmer';
        }
        else if (gentype == 'GENERIC_TYPE_SWITCH_BINARY') {
            devtype = 'Switch';
        }
        else {
            devtype = 'Unknown';
        }
        return devtype;
    };

    this.getDeviceByNode = function(nodeid) {
        for (d in this.devices) {
            if (nodeid == this.devices[d].node_id) {
            return this.devices[d];
            }
        }
        return null;
    };
   
    // 
    // render table of devices that we know about
    //
    this.renderDeviceTable = function(de) {
        if (this.devSorted.length > 0) {
        var ih = '<table class="devicetable">';
        ih += '<tr><th class="bull_col">&bull;</th><th class="indicator_col">&nbsp;</th><th class="idcol" id="idcol" onclick="setSortID();" title="sort by ID">ID</th><th class="typecol" id="typecol" onclick="setSortType();" title="sort by Type">Type</th><th class="namecol" id="namecol" onclick="setSortName();" title="sort by Name">Name</th></tr>';
        for (d in this.devSorted) {
            var ni = this.devSorted[d].node_id;
            // fetch prefs
            var dname = widget.preferenceForKey("dev." + ni + ".name");
            if (dname == undefined || dname.length==0) { 
                dname = '&lt;node ' + ni + '&gt;';
            } 
            // sort out checkbox...
            // not visible on front side 
            // not available for front side 
            // if is a controller
            var include_cb;
            if (this.devSorted[d].isMyController()) {
                 include_cb = '&nbsp;';
                 this.devSorted[d].front = false;
            }
            else {
                var ens='';
                if (this.devSorted[d].front)  {
                    ens = "checked";
                }
                include_cb = '<input type="checkbox"' + ens + ' onchange="devcheckhandler(this, '+ni+');">';
            }
            var indicate = '&nbsp;';
            if (this.devSorted[d].canIndicate()) {
                indicate = '<img src="Images/ktip.png" onclick="handleIndicator(this,' + ni + ');" title="Make physical device flash">';
            }
            ih += '<tr class="devicetablerow"><td class="bull_col">' + include_cb + '</td><td class="indicator_col">' + indicate + '</td><td class="idcol">' + this.devSorted[d].node_id + '</td><td class="typecol">' + this.devSorted[d].deviceType + '<td class="namecol"  onclick="devnamechangehandler(this, '+ni+');">' + dname + '</td></tr>';
        }
        ih += '</table>';
        }
        else {
            var ih = "<br/><br/><br/><br/>No devices found.</br>";
        }
        de.innerHTML = ih;
        this.renderSortColumns();
    }; 
    
    this.renderSortColumns = function() {
        if (this.sortPlan && this.sortPlan != null) {
            var p = this.sortPlan;
            document.getElementById('idcol').style.textDecoration='none';
            document.getElementById('typecol').style.textDecoration='none';
            document.getElementById('namecol').style.textDecoration='none';
            
            if (p.substring(0,2) == 'id') {
                eln = 'idcol';
            }
            else if (p.substring(0,4) == 'type') {
                eln = 'typecol';
            }
            else if (p.substring(0,4) == 'name') {
                eln = 'namecol';
            }
            if (eln != null) {
                var el = document.getElementById(eln);
                if (el) {
                    el.style.textDecoration='underline';
                }
            }
        }
    
    };    
    
    this.getLog = function() {
        // fill in log from the device
        o = new Object();
        serverSendCatch('net','get_net_log', true, this.saveLog, o);
    };

    this.clearLog = function() {
        // fill in log from the device
        o = new Object();
        serverSendCatch('net','clear_net_log', true, this.saveLog, o);
    };

    
    this.saveLog = function(request) {
    
        if (request.status == 200 && request.responseText) {
            this.log = JSON.parse(request.responseText)
            // display it!
            var html='';
            for (l in this.log) {
                html += '<p>' + log[l] + '</p>';
            }
            var scroll = document.getElementById('netlogscrollarea');
            scroll.object.content.innerHTML = html;
            scroll.object.verticalScrollTo(9999);
            scroll.object.refresh();
        }
    
    };
    
    this.showErrorForNode = function(nodeid) {
        //
        // show the error tab for this node id
        // and select and display and all
        // so, we need to find if this node is displayed on front, and if
        // so, show the errors
        var dev = this.getDeviceByNode(nodeid);
        if (dev && dev.front) {
            // find this one indexed on front
            var frontIndex = this.getDeviceIndex(nodeid);
            if (frontIndex >= 0) {
                selectDeviceFromIndex(frontIndex);
                showDrawer();
                swapDrawerErrors(null);
            }
        }
    };
    
    this.doSort = function () {
        // produce a sorted list of devices
        var sorted = new Array();
        // populate new array
        for (d in this.devices) {
            sorted.push(this.devices[d]);
        }
        var fn=null;
        // set sort type
        switch(this.sortPlan) {
        case 'id':
            fn=sortID;
            break;
        case 'id-':
            fn =sortIDRev;
            break;
        case 'type':
            fn=sortType;
            break;
        case 'type-':
            fn=sortTypeRev;
            break;
        case 'name':
            fn=sortName;
            break;
        case 'name-':
            fn=sortNameRev;
            break;
        default:
            fn=null;
        }
        // sort
        if (fn != null) {
            sorted.sort(fn);
        }
        // save
        this.devSorted = sorted;
    };
    
    this.saveDevPrefs = function() {
        var myc = this.getMyController();
        if (myc) {
            setPreferenceForKey(myc.home_id, 'net.homeid');
        }
    };

    
}

function sortID(a,b) {
    return a.node_id > b.node_id;
}

function sortIDRev(a,b) {
    return a.node_id < b.node_id;
}

function sortType(a,b) {
    return a.deviceType > b.deviceType;
}

function sortTypeRev(a,b) {
    return a.deviceType < b.deviceType;
}

function sortName(a,b) {
    return a.name > b.name;

}

function sortNameRev(a,b) {
    return a.name < b.name;

}

function Device(deviceType, name, channels, hasEarthSurge, l, earth, surge, manuf_id, prod_type, lib_type, lib_version, lib_sub_version, home_id, node_id) {
    this.selected = true;
    this.deviceType = deviceType;
    this.name = widget.preferenceForKey("dev." + node_id + ".name");
    this.channels = channels;
    this.level = l;
    this.savedLevel = l;
    this.prev_level = -1;
    this.hasEarthSurge = hasEarthSurge;
    this.earth = earth;
    this.surge = surge;
    this.manuf_id = manuf_id;
    this.prod_type = prod_type;
    this.lib_type = lib_type;
    this.lib_version = lib_version;
    this.lib_sub_version = lib_sub_version;
    this.home_id = home_id;
    this.node_id = node_id;
    this.slider = null;
    this.lastLevelChange = 0;
    // configuration
    this.config = new Object();
    this.config.rules = [false, false, false, false, false, false];
    this.config.schedule = new Object();
    this.config.schedule.weekdays = new Object();
    this.config.schedule.weekdays.on = false;
    this.config.schedule.weekdays.ontime = '09:00';
    this.config.schedule.weekdays.off = false;
    this.config.schedule.weekdays.offtime = '17:00';
    this.config.schedule.weekends = new Object();
    this.config.schedule.weekends.on = false;
    this.config.schedule.weekends.ontime = '09:00';
    this.config.schedule.weekends.off = false;
    this.config.schedule.weekends.offtime = '17:00';
    //log
    this.log=[];
    this.comm_failure = false;

    this.saveConfig = function() {
        var jstring = JSON.stringify(this.config);
        //logging.info('saving node' + this.node_id + ' json: ' + jstring); 
        setPreferenceForKey(jstring, 'dev.' + this.node_id + '.config');
        // and to server
        o = new Object();
        o.config = this.config;
        o.nodeid = this.node_id;
        serverSend('dev', 'set_config', true, o);

    };

    this.clearNameAndFront = function() {
        setPreferenceForKey(null, 'dev.' + this.node_id + '.front');
        setPreferenceForKey(null, 'dev.' + this.node_id + '.name');
        this.front = false;
    };
    
    this.saveDevPrefs = function() {
        var f = 'no';
        if (this.front) {
            f = 'yes';
        }
        setPreferenceForKey(f, 'dev.' + this.node_id + '.front');
        setPreferenceForKey(this.name, 'dev.' + this.node_id + '.name');
    };
    
    this.recallDevPrefs = function() {
        this.name = widget.preferenceForKey('dev.' + this.node_id + '.name');
        this.front = (widget.preferenceForKey('dev.' + this.node_id + '.front') == 'yes');
    };
    

    this.newConfig = function() {
        this.config = new Object();
        this.config.rules = [false, false, false, false, false, false, false];
        this.config.schedule = new Object();
        this.config.schedule.weekdays = new Object();
        this.config.schedule.weekdays.on = false;
        this.config.schedule.weekdays.ontime = '09:00';
        this.config.schedule.weekdays.off = false;
        this.config.schedule.weekdays.offtime = '17:00';
        this.config.schedule.weekends = new Object();
        this.config.schedule.weekends.on = false;
        this.config.schedule.weekends.ontime = '09:00';
        this.config.schedule.weekends.off = false;
        this.config.schedule.weekends.offtime = '17:00';
    };

    this.getConfig = function() {
        var jstring = widget.preferenceForKey('dev.' + this.node_id + '.config');
        if (jstring) {
            this.config = JSON.parse(jstring);
        }
        else {
            this.newConfig();
        }
    };
    
    this.getConfig();


    this.getLog = function() {
        // fill in log from the device
        o = new Object();
        o.nodeid=this.node_id;
        serverSendCatch('dev','get_log', true, this.saveLog, o);
    };

    this.clearLog = function() {
        o = new Object();
        o.nodeid=this.node_id;
        serverSendCatch('dev','clear_log', true, this.saveLog, o);
    }
    
    this.saveLog = function(request) {
    
        if (request.status == 200 && request.responseText) {
            this.log = JSON.parse(request.responseText)
            // display it!
            var el=document.getElementById('errors-text');
            var html='';
            for (l in this.log) {
                html += '<p>' + log[l] + '</p>';
            }
            //el.innerHTML = html;
            var scroll = document.getElementById('scrollArea1');
            
            scroll.object.content.innerHTML = html;
            scroll.object.verticalScrollTo(9999);
            scroll.object.refresh();
                
        }
        else {
            alert("Error fetching data: HTTP status " + xmlRequest.status);
        }
    
    };
    
    this.updateLevelFromZWave = function(newLevel) {
        
        var now = (new Date()).getTime();
        if ((now - this.lastLevelChange) > 2000) {
            this.prev_level = this.level;
            this.level = newLevel;
        }
    };


    this.isValidTime = function(el) {
        // valid patterns are:
        //hh:mm
        //h:mm
        //hhmm
        var s = el.value;
        var p = s.indexOf(':');
        if (p<0 && s.indexOf('.') > 0) {
            p = s.indexOf('.');
        }
        if (p >0) {
            var hrs = s.substring(0,p);
            var mins = s.substring(p+1);
        }
        else if (p < 0 && s.length == 4) {
            var hrs = s.substring(0,2);
            var mins = s.substring(2,4);
        }
        else {
            return false;
        }
        var h = parseInt(hrs,10);
        var m = parseInt(mins,10);
        if (isNaN(h) || isNaN(m)) {
            return false;
        }
        if (h < 0 || h > 23) {
            return false;
        }
        if (m < 0 || m > 59) {
            return false;
        }
        // rewrite in a valid form:
        var nv = '';
        if (h < 10) { 
            nv = '0'
        }
        nv += h.toString();
        nv += ':';
        if (m < 10) {
            nv += '0';
        }
        nv += m.toString();
        el.value=nv;
        return true;
    };
        


    this.timeFieldValidate = function(el_name) {
        var el=document.getElementById(el_name);
        if (el) {
            if (!this.isValidTime(el)) {
                el.style.backgroundColor="pink";
                return false;
            }
            else {
                el.style.backgroundColor='white';
                return true;
            }
        }
        else {
            return false;
        }
    };

    this.validateSchedule= function() {
        var v = true;
        v &= this.timeFieldValidate('weekdays-on-time');
        v &= this.timeFieldValidate('weekdays-off-time');
        v &= this.timeFieldValidate('weekends-on-time');
        v &= this.timeFieldValidate('weekends-off-time');
        return v;
    };


    this.updateSchedule = function() {

        if (!this.validateSchedule()) {
            return;
        }

        this.config.schedule.weekdays.on = document.getElementById('weekdays-on').checked;
        this.config.schedule.weekdays.ontime = document.getElementById('weekdays-on-time').value;
        this.config.schedule.weekdays.off = document.getElementById('weekdays-off').checked;
        this.config.schedule.weekdays.offtime = document.getElementById('weekdays-off-time').value;

        this.config.schedule.weekends.on = document.getElementById('weekends-on').checked;
        this.config.schedule.weekends.ontime = document.getElementById('weekends-on-time').value;
        this.config.schedule.weekends.off = document.getElementById('weekends-off').checked;
        this.config.schedule.weekends.offtime = document.getElementById('weekends-on-time').value;
        this.saveConfig();
    };
    
    
    this.renderSchedule = function() {

        document.getElementById('weekdays-on').checked = this.config.schedule.weekdays.on; 
        document.getElementById('weekdays-on-time').value = this.config.schedule.weekdays.ontime;
        document.getElementById('weekdays-off').checked = this.config.schedule.weekdays.off;
        document.getElementById('weekdays-off-time').value = this.config.schedule.weekdays.offtime;

        document.getElementById('weekends-on').checked = this.config.schedule.weekends.on;
        document.getElementById('weekends-on-time').value = this.config.schedule.weekends.ontime;
        document.getElementById('weekends-off').checked = this.config.schedule.weekends.off;
        document.getElementById('weekends-on-time').value = this.config.schedule.weekends.offtime;
    };
    
    
    this.updateRules = function() {
        for (var i=0;i<7;i++) {
            this.config.rules[i] = document.getElementById('rule' + i).checked;
        }
        this.saveConfig();
    };

    this.renderRules = function() {
        for (var i=0;i<7;i++) {
            document.getElementById('rule' + i).checked = this.config.rules[i];
        }    
    };

    this.renderKeyPairs = function(el) {
        var html = '<table class="drawer-details">';
        
        if (el) {
            html += this.addKRow('Manufacturer', false);
            html += this.addKRow('Product Type', false);
            html += this.addKRow('Product Id', false);
            html += this.addKRow('Base Class', false);
            html += this.addKRow('Generic Class', false);
            html += this.addKRow('Specific Class', false);
            html += this.addKRow('Cmd. Classes', true);            
            html += this.addKRow('Controller', true);
        }
        html += '</table>';
        el.object.content.innerHTML = html;
        el.object.verticalScrollTo(0);
        el.object.refresh();
    };
     
    this.addKRow = function(key, repl) {
        var h = '';
        if (this.details && this.details[key]) {
            var da = this.details[key];
            if (repl) {
                da = da.replace(/,/g,'<br/>');
            }
            h = '<tr class="draw-det-row"><td><strong>' + key + ':&nbsp;</strong></td><td>' + da + '</td></tr>';
        }
        return h;
    };


    this.renderUpdate = function(element, index) {
        // see if we have what we need
        var dt = document.getElementById('deviceType' + index);
        if (dt) {
            // fill in existing divs
            dt.innerHTML = this.deviceType;
            if (this.channels == 2) {
                var dn1 = document.getElementById('deviceName1-' + index);
                dn1.innerHTML = this.name[0];
                var dn2 = document.getElementById('deviceName2-' + index);
                dn2.innerHTML = this.name[1];
            }
            else {
                // 1 channel
                // name
                var dn = document.getElementById('deviceName-' + index);
                var n = this.name;
                if (n == undefined || n.length == 0) { 
                    n = '&lt;node ' + this.node_id + '&gt;';
                }                   
                dn.innerHTML = n;
                var ooc = document.getElementById('onOffControl-' + index);
                // On Off control
                // 1 channel
                var onclass="on";
                var offclass = "off";
                if (this.comm_failure) {
                    onclass="vflash";
                    offclass="vflash";
                }                
                if (ooc) {
                    var l = '<span class="' + offclass + '" onclick="toggle(event, ' + index + ',0)" title="Click to turn On">Off</span>';
                    if (this.level > 0) { 
                        l = '<span class="' + onclass + '" onclick="toggle(event, ' + index + ',0)" title="Click to turn Off">On</span>';
                    }
                    if (this.deviceType == 'Switch') {
                        ooc.innerHTML = l;
                    }
                    else {
                        if (this.level > 0) {
                            ooc.innerHTML = '<span class="' + onclass + '" onclick="toggle(event, ' + index + ',0)" id="percent' + index + '" title="Click to turn Off">' + this.level +'%</span>';
                        }
                        else {
                            ooc.innerHTML = '<span class="'+ offclass + '" onclick="toggle(event, ' + index + ',0)" id="percent' + index + '" title="Click to turn On">' + this.level +'%</span>';                        }
                        
                        // slider update
                        this.renderSliders(index);
                        //if (this.slider) {
                        //    this.slider.setValue(this.level/100);
                        //}

                    }
                }
            }        
        }
        else {
            // zero out slider, coz we can't find its place (where has it gone?)
           // this.slider = null;
            this.newRenderInto(element, index);
        }
    
    
    
    };
    
    
    this.newRenderInto = function(element, index) {
        var myel = document.getElementById('box' + index);
        var dt = document.createElement('div');
        dt.className = 'deviceType';
        dt.id='deviceType' + index;
        dt.setAttribute('id','deviceType' + index);
        dt.setAttribute('name','deviceType' + index);
        dt.onclick = selectDevice;
        dt.innerHTML = this.deviceType;
        var dn = document.createElement('div');
        dn.className = 'deviceName';
        dn.setAttribute('id','deviceName-' + index);
        dn.id = 'deviceName-' + index;
        dn.onclick = selectDevice;
        var n = this.name;
        if (n == undefined || n.length==0) { 
            n = '&lt;node ' + this.node_id + '&gt;';
        } 
        dn.innerHTML=n;

        var ctrl = document.createElement('div');
        ctrl.className = 'onOffControl';
        ctrl.setAttribute('id', 'onOffControl-' + index);
        ctrl.onclick = selectDevice;
        ctrl.innerHTML = this.renderControls(index);
        var sl = document.createElement('div');
        sl.className = 'slider';
        sl.setAttribute('id', 'slider' + index);
        // add these elements
        myel.insertBefore(dt,null);
        myel.insertBefore(dn,null);
        myel.insertBefore(ctrl,null);
        myel.insertBefore(sl, null);
        this.renderSliders(index);
        if (this.deviceType=='Dimmer' && this.level != this.prev_level) {
        //    this.setDimLevel(this.level/100, index);
            this.prev_level = this.level;
        }
    };

    this.renderInto = function(element, index) {
        var html = '<div class="deviceType" id="deviceType' + index + '">' + this.deviceType + '</div>';
        if (this.channels == 2) {
            html += '<div class="deviceName1" id="deviceName1-' + index + '">' + this.name[0] + '</div>';
            html += '<div class="deviceName2" id="deviceName2-' + index + '">' + this.name[1] + '</div>';
        }
        else {
            var n = this.name;
            if (n == undefined || n.length == 0) { 
                n = '&lt;node ' + this.node_id + '&gt;';
            } 
            
            html += '<div class="deviceName" id="deviceName-' + index + '">' + n + '</div>';        
        }
                            // now control bits
        html += this.renderControls(index);
        html += this.renderEarthSurge(index);
        element.innerHTML = html;
        this.renderSliders(index);
        if (this.deviceType=='Dimmer' && this.level != this.prev_level) {
            //this.setDimLevel(this.level/100, index);
            this.prev_level = this.level;
        }
    };
    
    //
    // renders the On Off bit
    //
    this.renderControls = function(index) {
        if (this.channels == 2) {
            if (this.deviceType=='Switch') {
                var l = new Array();
                for (i in this.level) {
                    var r = i + 1;
                    if (this.level[i] > 0) {
                        l[i] = '<span class="on" onclick="toggle(event, '+ index + ',' + r + ')" title="Click to turn Off">On</span>';
                    }
                    else {
                        l[i] = '<span class="off" onclick="toggle(event, '+ index + ',' + r + ')" title="Click to Turn On">Off</span>';
                    }
                }
                return '<div class="onOffControl1" id="onOffControl1-' + index + '">' + l[0] + '</div>' + '<div class="onOffControl2" id="onOffControl2-' + index +'">' + l[1] + '</div>';
            }
        }
        else if (this.deviceType=='Switch' || this.deviceType == 'Dimmer'){
            // 1 channel
            var onclass="on";
            var offclass = "off";
            if (this.comm_failure) {
                var onclass="vflash";
                var offclass="vflash";
            }
            var l = '<span class="' + offclass + '" onclick="toggle(event, ' + index + ',0)" title="Click to turn Off">Off</span>';
            if (this.level > 0) { 
                l = '<span class="' + onclass + '" onclick="toggle(event, ' + index + ',0)" title="Click to turn On">On</span>';
            }
            if (this.deviceType == 'Switch') {
                return l;
            }
            else {
                if (this.level > 0) {
                    return '<span class="'+ onclass + '" onclick="toggle(event, ' + index + ',0)" id="percent' + index + '" title="Click to turn Off">' + this.level +'%</span>';
                }
                else {
                    return '<span class="' + offclass + '" onclick="toggle(event, ' + index + ',0)" id="percent' + index + '" title="Click to turn On">' + this.level +'%</span>';
                }
            }
        }
    };
        
    
    this.renderSliders = function(index) {
    
        if (this.deviceType == 'Dimmer') {
            if (this.slider == null) {
                this.slider = new AppleHorizontalSlider(document.getElementById("slider" + index), null);
                this.slider.continuous = false;
                this.slider.setValue(this.level/100);
                this.slider.refresh();
                this.slider.onchanged = function(val) { if (zwave) { var dev=zwave.getFrontDevice(index); if (dev) {  dev.handleDimLevelChange(val, index);} } };
            }
            if (this.slider != null) {
                if (this.slider._thumbStartTemp == undefined) {
                this.slider.setValue(this.level/100);
                this.slider.refresh();                    
                }
                this.slider.onchanged = function(val) { if (zwave) { var dev=zwave.getFrontDevice(index); if (dev) { dev.handleDimLevelChange(val, index); } } };
            }
            var e = document.getElementById('slider' + index);
            if (e) {
                if (this.comm_failure) {
                    e.style.display="none";
                }
                else {
                    e.style.display="block";
                }
            }
        }
    };
    
    this.renderEarthSurge = function() {
        var html='';
        if (hasEarthSurge) {
            if (surge > 0) {
                html = '<div class="surge good">Surge</div>';
            }
            else {
                html = '<div class="surge bad">Surge</div>';            
            }
            if (earth > 0) {
                html += '<div class="earth good">Earth</div>';
            }
            else {
                html += '<div class="earth bad">Earth</div>';            
            }
        }
        return html;
    
    };
    
    this.handleDimLevelChange = function(xl, index) {
        //logging.info('new level from slider is: '+ xl);
        if (xl == null || isNaN(xl)) {
            return;
        }
        // level is floating point 0..1 so get to 0..100
        var lev = Math.round(xl*100);
        this.lastLevelChange = (new Date()).getTime();
        this.prev_level = this.level;
        this.level = lev;
        this.setDimLevel(lev, index);
        var o = Object();
        o.nodeid = this.node_id;
        o.cclass = 'switch_multilevel';
        o.ccmd = 'level_set';
        o.level = lev;
        serverSend('dev', 'ccmd', true, o);
    };
    
    this.setDimLevel = function(l, index) {
        this.prev_level = this.level;
        this.level = l;
        var percent = document.getElementById('percent' + index);
        percent.innerHTML = Math.round(l) + '%';
    };
    
    this.toggle = function(ch) {
        if (this.comm_failure) {
            return;
        }
        if (this.channels == 1) {
            if (this.deviceType == 'Switch') {
                if (this.level > 0) {
                    this.level = 0;
                }
                else {
                    this.level = 100;
                }
                this.lastLevelChange = (new Date()).getTime();
                var o = Object();
                o.nodeid = this.node_id;
                o.cclass = 'switch_binary';
                o.ccmd = 'level_set';
                o.level = this.level;
                serverSend('dev', 'ccmd', true, o);
            }
            else if (this.deviceType == 'Dimmer') {
                // dimmer with saved levels
                if (this.level > 0) {
                    this.savedLevel = this.level;
                    this.level = 0;
                }
                else {
                    this.level = this.savedLevel;
                }
                this.lastLevelChange = (new Date()).getTime();
                var o = Object();
                o.nodeid = this.node_id;
                o.cclass = 'switch_multilevel';
                o.ccmd = 'level_set';
                o.level = this.level;
                serverSend('dev', 'ccmd', true, o);
            }
            else { // Unknown/other devices
                if (this.level > 0) {
                    this.savedLevel = this.level;
                    this.level = 0;
                }
                else {
                    this.level = this.savedLevel;
                }
                this.lastLevelChange = (new Date()).getTime();
                var o = Object();
                o.nodeid = this.node_id;
                o.cclass = 'basic';
                o.ccmd = 'level_set';
                o.level = this.level;
                serverSend('dev', 'ccmd', true, o);
            }

        }
        else {
            if (this.deviceType == 'Switch') {
                if (this.level[ch-1] > 0) {
                    this.level[ch-1] = 0;
                }
                else {
                    this.level[ch-1] = 100;
                }
            }
            else {
                // dimmer with saved levels
                if (this.level[ch-1] > 0) {
                    this.savedLevel[ch-1] = this.level[ch-1];
                    this.level[ch-1] = 0;
                }
                else {
                    this.level[ch-1] = this.savedLevel[ch-1];
                }
            }
        }
    };
  
    // check to see if this device is the controller
    this.isMyController = function() {
        return (this.controller_capabilities != null);
    };
  
    // check to see that this device is a secondary controller
    this.isSecondary = function() {
        return (this.isMyController() && (this.controller_capabilities.search('secondary') != -1));
    };
    
    // check to see that this device is a secondary controller
    this.isInclusion = function() {
        return (this.isMyController() && (this.controller_capabilities.search('inclusion') != -1));
    };
    
    this.isPrimary = function() {
        return (this.isMyController() && (this.controller_capabilities.search('primary') != -1));
    };
    
    this.setName = function(name) {
        this.name = name;
        setPreferenceForKey(name, "dev." + this.node_id + ".name");
    };
    
    this.canIndicate = function () {
        //TODO perform proper test here
        return !this.isMyController();
    };
          
    this.indicate = function() {
        // send indicate command to plugin, to flash something
        var o = Object();
        o.nodeid = this.node_id;
        serverSend('dev', 'indicate', true, o);
    };
    
    this.is_tricklestar = function() {
        return (this.details && this.details['Manufacturer'] == 'TrickleStar');
    }
    
}

// handlers

// replace the name with a text box to type a name into
// with a handler
function devnamechangehandler(el, ni) {
    zwave.enableDeviceTable = false;
    // retrieve existing name for text field
    var existName = el.innerText;
    el.onclick=null;
    el.innerHTML='<input id="devname' + ni + '" type="text" size=14 maxlength="14" onblur="saveDevName(this, event,' + ni + ');" onkeypress="return devNamePress(this, event, ' + ni + ');" value="' + existName +'">';
    var tf = document.getElementById('devname' + ni);
    tf.focus();
}


function devNamePress(el, event, ni) {
    if (isReturn(event)) {
        saveDevName(el,event,ni);
        return false;
    }
    else {
        return true;
    }

}


function isReturn(event) {
    var keynum = event.which;
    return (keynum == 0x0d);
}

function saveDevName(el, event, ni) {
    var dev = zwave.getDeviceByNode(ni);
    if (dev) {
        dev.setName(el.value);
    }
    zwave.enableDeviceTable = true;
    zwave.doSort();
    
}


// name has been changed, so save it and replace the 
function devnamechangedhandler(el, ni) {

    zwave.enableDeviceTable = true;
    zwave.doSort();
}

// indicator support
function handleIndicator(el, ni) {
    var dev = zwave.getDeviceByNode(ni);
    if (dev) {
        dev.indicate();
    }

}

// indicator support
function handleClearLog(event) {
    if (selected >=1 && selected <=8) {
        var dev = zwave.getFrontDevice(selected);
        if (dev) {
            dev.clearLog();
        }
    }
}

// indicator support
function handleRefreshLog(event) {
    if (selected >=1 && selected <=8) {
        var dev = zwave.getFrontDevice(selected);
        if (dev) {
            dev.getLog();
        }
    }
}

function setSortID(event) {
    setSortPlan('id');
}

function setSortType(event) {
    setSortPlan('type');
}

function setSortName(event) {
    setSortPlan('name');
}

function setSortPlan(plan) {
    if (zwave) {
        var currentPlan = zwave.sortPlan;
        if (currentPlan != plan) {
            zwave.sortPlan = plan;
        } 
        else {
            zwave.sortPlan = plan + '-';
        }
        zwave.clearFrontSide();
        zwave.doSort();
        zwave.renderDeviceTable(document.getElementById('box-content'));
    }
}
